import argparse
import os
import sys
import re
import subprocess
import time
from base64 import b64encode
from binascii import unhexlify, hexlify
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
import string
import random
from dotenv.main import load_dotenv

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
BASE_DIR = os.path.join(FILE_DIR, "..")
sys.path.append(BASE_DIR)

from drippy.drip_iroha_account import DripIrohaAccount
from drippy.gps_poll import GpsPoller
from drippy.util import astm_float_to_int_hex, get_timestamp, astm_time
from drippy.host_identity import HostIdentity

# File with flight_id that will be incremented and used if no other
# flight_id is specified.
FLIGHT_ID_FILE = os.path.join(FILE_DIR, "default_flight_id.txt")


load_dotenv()


def separate_bytes(hex_str, delim=' '):
    return re.sub(r'.{2}', r'\g<0>{}'.format(delim), hex_str).strip()


AUTH_PAGE_0_LEN = 17
AUTH_PAGE_LEN = 23

g_lat, g_lon = 58.398176, 15.575927

executor = ThreadPoolExecutor(5)


def get_auth_pages(signature: bytes, auth_type=0x10):
    """
    Build a list of ASTM F3411-19 authentication pages with the
    signature provided.
    """
    auth_pages = []
    auth_len = len(signature)

    # Pad signature with 0s so all pages are even length.
    signature += b'\x00' * \
        (AUTH_PAGE_LEN - ((len(signature)-AUTH_PAGE_0_LEN) % AUTH_PAGE_LEN))
    pages = 1 + (len(signature) - AUTH_PAGE_0_LEN)//AUTH_PAGE_LEN

    # First page is 7 bytes header + 17 bytes from signature
    auth_page_0 = b'\x10' + pages.to_bytes(1, 'big') + auth_len.to_bytes(
        1, 'big') + int(astm_time()).to_bytes(4, 'little')
    auth_page_0 += signature[:AUTH_PAGE_0_LEN]
    auth_pages.append(auth_page_0)

    # Remaining pages are 1 byte header + 23 bytes from signature
    i = AUTH_PAGE_0_LEN
    page = 1
    bytes_to_encode = len(signature) - AUTH_PAGE_0_LEN
    while i < bytes_to_encode:
        auth_page = (auth_type | page).to_bytes(1, 'big')
        auth_page += signature[i:i+AUTH_PAGE_LEN]
        auth_pages.append(auth_page)

        i += AUTH_PAGE_LEN
        page += 1

    return auth_pages


def location_update(gpsd: GpsPoller, iroha: DripIrohaAccount, flight_id):
    # ASTM F3-411 Location message 0x1
    print("location_update")
    global executor

    # If gpsd is active, check that it has received values
    if gpsd and gpsd.get_current_value() == None:
        return

    if gpsd:
        lat = separate_bytes(astm_float_to_int_hex(gpsd.get_lat()))
        lon = separate_bytes(astm_float_to_int_hex(gpsd.get_lon()))
    else:
        lat = separate_bytes(astm_float_to_int_hex(g_lat))
        lon = separate_bytes(astm_float_to_int_hex(g_lon))

    payload = "10 00 00 00 {lat} {lon} 00 00 00 00 00 00 00 00 {ts} 00 00".format(
        lat=lat, lon=lon, ts="00 00")

    if iroha:  # if iroha private supplied, send location transactions on separate thread.
        # location will be flight id concatenated with the bluetooth payload
        # In order to save space we can b64encode the bytes of the payload,
        # instead of sending a hexstring which would result in location message
        # taking up double the bytes (2 characters per byte).
        iroha_msg = b64encode(
            unhexlify(flight_id + ''.join(payload.split(' ')))).decode("utf-8")
        executor.submit(
            iroha.set_account_details, "location", iroha_msg
        )

    subprocess.run(
        "hcitool -i hci0 cmd 0x08 0x0008 1f 1e 16 fa ff 0d 00 10 {payload}".format(payload=payload), shell=True)


def basic_id_update(hhit):
    # ASTM F3-411 Basic ID message 0x0 with HHIT.
    print("basic_id_update")
    subprocess.run(
        "hcitool -i hci0 cmd 0x08 0x0008 1f 1e 16 fa ff 0d 00 00 42 {} 00 00 00 00 00 00 00".format(hhit), shell=True)
    time.sleep(0.5)


def auth_update(auth_pages):
    print("auth_update")
    for page in auth_pages:
        msg = separate_bytes(hexlify(page).decode("utf-8"))
        subprocess.run(
            "hcitool -i hci0 cmd 0x08 0x0008 1f 1e 16 fa ff 0d {} 20 {}".format("00", msg), shell=True)
        time.sleep(0.2)


DESCRIPTION = """Send ble advertisements using the specified Host Identity,
in accordance to ASTM F3411-19.

If iroha option with the private key file is specified,
location updates will be sent to Iroha blockchain as well.
The private key should belong to the same host specificed by
the host identity.

If no flight_id is specified, then a random flight id will be generated.
flight_id is only used if the --iroha option is specified.
"""


def main():
    global g_lat
    global g_lon
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        "-f", "--file",
        dest="host_identity_file",
        default="/usr/local/etc/hip/my_host_identities.xml",
        type=str,
        help="Specify the host identity xml file."
    )
    parser.add_argument(
        "-i", "--iroha",
        dest="iroha_priv_key_file",
        default=None,
        type=str,
        help="Specify the iroha private key file to enable location updates to Iroha."
    )
    parser.add_argument(
        "--gps",
        help="Turn on gps updates",
        action="store_true"
    )
    parser.add_argument(
        "--flight_id",
        type=int,
        help="Specify a flight_id for the location updates. Used to identify the flight path in the blockchain. This value is only used if the --iroha option is specified."
    )
    parser.add_argument(
        "--run-scenario",
        dest="run_scenario",
        action="store_true",
        help="Run a flight scenario with randomized coordinates originating at ({}, {}). This will disregard the gps option.".format(
            g_lat, g_lon)
    )
    parser.add_argument(
        "--expiration-date",
        dest="exp_date",
        default="2030-01-01T13:00",
        type=str,
        help="Specify expiration date for the attestation in the format YYYY-mm-ddTHH:MM"
    )

    args = parser.parse_args(sys.argv[1:])
    hi_file = args.host_identity_file
    iroha_priv_key_file = args.iroha_priv_key_file

    exp_date = datetime.strptime(args.exp_date, "%Y-%m-%dT%H:%M")
    if (exp_date - datetime.utcnow()).total_seconds() <= 0:
        print("Expiration date cannot come before current date.")
        sys.exit(-1)

    if not os.path.isfile(hi_file) \
            or hi_file[-3:] != "xml":
        print("The specified file is not an xml file")
        sys.exit(-1)

    hi = HostIdentity.from_file(hi_file)

    iroha = None
    if iroha_priv_key_file:
        if not os.path.isfile(iroha_priv_key_file) \
                or iroha_priv_key_file[-4:] != "priv":
            print("The specified iroha private key file is not a .priv file")
            sys.exit(-1)

        key = open(iroha_priv_key_file, 'r').read()
        iroha = DripIrohaAccount(key, hi.get_hhit(), hi.get_hid(
        ), os.environ.get("IROHA_HOST"), os.environ.get("IROHA_PORT"))

    if args.flight_id:
        flight_id = args.flight_id
    else:
        flight_id = ''.join(random.SystemRandom().choice(
            string.hexdigits) for _ in range(64))

    hhit = separate_bytes(hi.get_hhit())
    attest = hi.generate_self_concise_attestation(get_timestamp(exp_date))
    if attest is None:
        print("Expiration date has expired.")
        sys.exit(-1)

    auth_pages = get_auth_pages(attest)

    # Reset bt
    subprocess.run("hcitool -i hci0 cmd 0x03 0x0003", shell=True)

    # AD parameters, 100 ms interval on all channels.
    subprocess.run(
        "hcitool -i hci0 cmd 0x08 0x0006 a0 00 a0 00 03 00 00 00 00 00 00 00 00 07 02", shell=True)

    # Enable bluetooth advertisement.
    subprocess.run("hcitool -i hci0 cmd 0x08 0x000a 01", shell=True)

    gpsd = None
    if not args.run_scenario and args.gps:
        gpsd = GpsPoller()
        gpsd.start()

    i = 0
    try:
        while True:
            os.system('clear')
            print("The flight_id is {}".format(flight_id))

            # Send basic_id and auth at least once every 3 seconds.
            # Send basic_id first second
            if i % 3 == 0:
                basic_id_update(hhit)
            # Send first half of auth pages second second
            elif i % 3 == 1:
                auth_update(auth_pages[:2])
            # Send last half of auth pages last second
            elif i % 3 == 2:
                auth_update(auth_pages[2:])

            # Location updates at least once every second.
            location_update(gpsd, iroha, flight_id)

            if args.run_scenario:
                g_lat += random.uniform(0.0001, 0.001) * \
                    random.choice([-1, 0, 1])
                g_lon += random.uniform(0.0001, 0.001) * \
                    random.choice([-1, 0, 1])

            i += 1
            time.sleep(1 - time.monotonic() % 1)
    except KeyboardInterrupt:
        # Reset bt
        subprocess.run("hcitool -i hci0 cmd 0x03 0x0003", shell=True)


if __name__ == '__main__':
    main()
