import binascii
from iroha import IrohaCrypto
from iroha import Iroha, IrohaGrpc

from .host_identity import HostIdentity
from .util import int_bytes_to_astm_float

from concurrent.futures import ThreadPoolExecutor
from google.protobuf.json_format import MessageToDict

import json
import re
import sys
from base64 import b64decode
from binascii import hexlify, unhexlify

FLIGHT_ID_SIZE = 32

def int_bytes_to_astm_float(int_hex: bytes):
  return int.from_bytes(int_hex, 'little')/1e7

def parse_b64location(b64location):
  try:
    loc = b64decode(b64location)
    loc_flight_id = loc[:FLIGHT_ID_SIZE]
    payload = loc[FLIGHT_ID_SIZE:]
    lat = int_bytes_to_astm_float(payload[4:8])
    lng = int_bytes_to_astm_float(payload[8:12])

    return {
      "flight_id": hexlify(loc_flight_id).decode("utf-8"),
      "lat": lat,
      "lng": lng
    }
  except binascii.Error:
    return None


class GeoPos:
  def __init__(self, lat, lng, time):
    self._lat = lat
    self._lng = lng
    self._time = time

  def get_lat(self):
    return self._lat

  def get_lng(self):
    return self._lng

  def get_geo_position(self):
    return (self._lat, self._lng)

  def get_time(self):
    return self._time

  def __str__(self):
    return "({}, {}) @ {}".format(self._lat, self._lng, self._time)


class DripIrohaAccount:
  def __init__(self, iroha_priv_key, hit, hid, host, port):
    self.account_id = "{}@{}".format(hit, hid)
    self.iroha = Iroha(self.account_id)
    self.priv_key = iroha_priv_key

    self.net = IrohaGrpc('{}:{}'.format(host, port))

  def get_account_id(self):
    return self.account_id


  def set_account_details(self, key, val, account_id=None):
    tx = self.iroha.transaction([
      self.iroha.command(
        "SetAccountDetail",
        account_id=account_id if account_id else self.account_id,
        key=key,
        value=val
      )
    ])
    IrohaCrypto.sign_transaction(tx, self.priv_key)
    self.net.send_tx(tx)

    for status in self.net.tx_status_stream(tx):
      print(status)


  def get_account(self, account_id):
    query = self.iroha.query("GetAccount", account_id=account_id)
    IrohaCrypto.sign_query(query, self.priv_key)

    response = self.net.send_query(query)
    return json.loads(response.account_response.account.json_data)


  def get_blocks(self):
    query = self.iroha.blocks_query()
    IrohaCrypto.sign_query(query, self.priv_key)
    blocks = self.net.send_blocks_stream_query(query, timeout=120)

    print(next(blocks))

  def get_block(self, height):
    query = self.iroha.query("GetBlock", height=height)
    IrohaCrypto.sign_query(query, self.priv_key)

    response = MessageToDict(self.net.send_query(query))
    if response.get("errorResponse") != None:
      print("Reason: {}\nMessage: {}".format(response["errorResponse"]["reason"], response["errorResponse"]["message"]))
      return False, response["errorResponse"]["message"]
    else:
      return True, response["blockResponse"]["block"]["blockV1"]["payload"]


  def find_block(self, height):
    query = self.iroha.query("GetBlock", height=height)
    IrohaCrypto.sign_query(query, self.priv_key)

    response = MessageToDict(self.net.send_query(query))
    if response.get("errorResponse") != None:
      print("Reason: {}\nMessage: {}".format(response["errorResponse"]["reason"], response["errorResponse"]["message"]))
      return None
    else:
      return response["blockResponse"]["block"]["blockV1"]["payload"]


  def find_geo_pos(self, height, account_id, flight_id):
    """
    Helper function for finding blocks using threads.

    @param geo_locations: global list in main thread,
    where worker threads will append geo locations.
    """

    query = self.iroha.query("GetBlock", height=height)
    IrohaCrypto.sign_query(query, self.priv_key)

    response = MessageToDict(self.net.send_query(query))
    if response.get("errorResponse") != None:
      return
    else:
      block = response["blockResponse"]["block"]["blockV1"]["payload"]

    try:
      for tx in block["transactions"]:
        for command in tx["payload"]["reducedPayload"]["commands"]:
          if not command.get("setAccountDetail"):
            continue

          cmd = command["setAccountDetail"]
          if not (cmd["key"] == "location" and cmd["accountId"] == account_id):
            continue

          location = parse_b64location(cmd["value"])
          if location and unhexlify(location["flight_id"]) == unhexlify(flight_id):
            # TODO: Get all values from the payload
            lat = location["lat"]
            lng = location["lng"]
            return GeoPos(lat, lng, tx["payload"]["reducedPayload"]["createdTime"])
    except KeyError:
      return None
    return None


  def find_geo_path(self, account_id, flight_id):
    """
    Return a list of all geo location updates for a given account,
    and flight id if specified.
    """

    # Ugly hack to get current block size of blockchain:
    # The error message of a GetBlock query will tell us the current size
    # of the blockchain. To get the size, just try query for a block
    # that does not exist.
    height = int(re.findall(r'(?<=\()\d+(?=\))', self.get_block(sys.maxsize)[1])[1])

    # get all geo locations using threads.
    geo_locations = []
    futures = []
    with ThreadPoolExecutor(max_workers=100) as executor:
      for i in range(1, height+1):
        futures.append(executor.submit(self.find_geo_pos, i, account_id, flight_id))

    for future in futures:
      result = future.result()
      if result:
        geo_locations.append(result)

    #geo_locations.sort(key=lambda loc: loc.get_time())

    return geo_locations