import sys
from binascii import hexlify, unhexlify

from bs4 import BeautifulSoup
from nacl.signing import SigningKey

from .util import *
from .cSHAKE import cSHAKE128


CONTEXT_ID = "F0EFF02FBFF43D0FE7930C3C6E6174EA"
CURVE_ID = "0001"

class HostIdentity:
  def __init__(self, hhit, hid, priv_key, pub_key):
    self._hhit = hhit
    self._hid = hid
    self._priv_key = priv_key
    self._pub_key = pub_key

    if not self.verify_hhit():
      print("The public key does not produce the same hash in hhit.")
    else:
      print("hhit verified")

  @staticmethod
  def from_file(file_name):
    bs = None
    try:
      with open(file_name, 'r') as f:
        bs = BeautifulSoup(f.read(), 'lxml')
    except PermissionError as e:
      print(e)
      sys.exit(1)

    host_id = bs.find("host_identity") # use first host id
    priv_key = ''.join(host_id.find("priv").string.split(':'))
    pub_key = ''.join(host_id.find("pub").string.split(':'))
    hid = host_id.find("hid").string

    hhit = ''.join(["{:>04}".format(byte_group) for byte_group in host_id.find("hit").string.split(':')])

    return HostIdentity(hhit, hid, priv_key, pub_key)

  def get_hhit(self):
    return self._hhit

  def get_hid(self):
    return self._hid

  def get_pub_key(self):
    return self._pub_key

  def get_priv_key(self):
    return self._priv_key


  def verify_hhit(self):
    chash = cSHAKE128(
      unhexlify(CONTEXT_ID + CURVE_ID + self._pub_key),
      64,
      "",
      unhexlify(CONTEXT_ID))

    return chash.decode("utf-8") == self._hhit[16:]


  def generate_self_attestation(self, expiration):
    if expiration <= astm_time():
      print("expiration timestamp has already expired.")
      return None

    message = self._hhit + self._pub_key + "{:>08x}".format(convert_be_to_le(expiration))
    sign_key = SigningKey(unhexlify(self._priv_key))
    return hexlify(sign_key.sign(unhexlify(message)))


  def generate_self_concise_attestation(self, expiration):
    if expiration <= astm_time():
      print("expiration timestamp has already expired.")
      return None

    message = self._hhit + "{:>08x}".format(convert_be_to_le(expiration))
    sign_key = SigningKey(unhexlify(self._priv_key))
    return sign_key.sign(unhexlify(message))


  def generate_certificate(self, other, expiration):
    signing_ts = astm_time()
    if expiration <= signing_ts:
      print("expiration timestamp has already expired.")
      return None

    ass_x = self.generate_self_attestation(expiration).decode("utf-8")
    ass_y = other.generate_self_attestation(expiration).decode("utf-8")

    len_ass_x = hex(struct.unpack("!2H", struct.pack("!i", len(ass_x)))[1])[2:]
    while len(len_ass_x) < 4:
      len_ass_x = '0' + len_ass_x

    len_ass_y = hex(struct.unpack("!2H", struct.pack("!i", len(ass_y)))[1])[2:]
    while len(len_ass_y) < 4:
      len_ass_y = '0' + len_ass_y

    message = len_ass_x + len_ass_y \
      + ass_x + ass_y \
      + hexlify(int(signing_ts).to_bytes(4, 'little')).decode("utf-8") \
      + hexlify(int(expiration).to_bytes(4, 'little')).decode("utf-8")

    sign_key = SigningKey(unhexlify(self._priv_key))
    return hexlify(sign_key.sign(unhexlify(message)))




